from types import BuiltinFunctionType
from operator import add
from numpy import zeros, array
from collections import deque


class Path(object):

    _REF = add

    def __init__(self, G, path, max_res, min_res):
        self.G = G
        self.path = path
        self.max_res = max_res
        self.min_res = min_res

    def __repr__(self):
        return str(self)

    def __str__(self):  # for printing purposes
        return "Path({0},{1},{2})".format(self.path, self.max_res, self.min_res)

    def _check_feasibility(self):
        shortest_path_edges = deque(e for e in zip(self.path, self.path[1:]))
        if any(e not in self.G.edges() for e in shortest_path_edges):
            raise Exception("No edge found")
        else:
            shortest_path_edges_w_data = deque()
            for e in shortest_path_edges:
                edge = deque(edge for edge in self.G.edges(data=True)
                             if edge[0:2] == e)[0]
                shortest_path_edges_w_data.append(edge)

        total_res = zeros(self.G.graph['n_res'])
        # Check path for resource feasibility by adding one edge at a time
        for edge in shortest_path_edges_w_data:
            if isinstance(self._REF, BuiltinFunctionType):
                total_res += self._edge_extract(edge)
            else:
                total_res = self._REF(total_res, edge)
            if (all(total_res <= self.max_res) and
                    all(total_res >= self.min_res)):
                pass
            else:
                break
        else:
            return True
        return edge

    @staticmethod
    def _edge_extract(edge):
        return array(edge[2]['res_cost'])



from logging import getLogger
from networkx import single_source_bellman_ford
from cspy.checks import (_check_res, _check_direction, _check_graph_attr,
                         _check_edge_attr, _check_path, _check_REFs)

log = getLogger(__name__)


def check(G, max_res=None, min_res=None, direction=None, algorithm=None):
    """Check whether inputs have the appropriate attributes and
    are of the appropriate types."""

    errors = []
    # Select checks to perform based on the input provided
    if max_res and min_res and direction:
        check_funcs = [
            _check_res, _check_direction, _check_graph_attr, _check_edge_attr,
            _check_path
        ]
    elif max_res and min_res:
        check_funcs = [
            _check_res, _check_graph_attr, _check_edge_attr, _check_path
        ]
    else:
        check_funcs = [_check_path]
    # Check all functions in check_funcs
    for func in check_funcs:
        try:
            func(G, max_res, min_res, direction, algorithm)
        except Exception as e:
            errors.append(e)  # if check fails save error message
    if errors:
        # if any check has failed raise an exception with all the errors
        raise Exception('\n'.join('{}'.format(item) for item in errors))


def prune_graph(G, max_res, min_res):
    """Removes nodes that cannot be reached due to resource limits.

    Note
    -----
    path_s and path_t contains all partial paths e.g.
        {node_i: [Source, ..., node_k, node_i]}.
    or  {node_i: [Sink, ..., node_k, node_i]}
    Hence, if node_i is found to violate a resource bound, it is
    because of node_k, therefore, we add path[key][-2] = node_k to
    the dictionary of nodes to remove.
    """

    def _check_resource(r):
        # check resource r's feasibility along a path

        def __get_weight(i, j, attr_dict):
            # returns number to use as weight for the algorithm
            return attr_dict['res_cost'][r]

        # Get paths from source to all other nodes
        length_s, path_s = single_source_bellman_ford(G,
                                                      'Source',
                                                      weight=__get_weight)
        length_t, path_t = single_source_bellman_ford(G.reverse(copy=True),
                                                      'Sink',
                                                      weight=__get_weight)
        try:
            # Collect nodes in paths that violate the resource bounds
            # see note above
            nodes_source.update({
                path_s[key][-2]: (val, r)
                for key, val in length_s.items()
                if val > max_res[r] or val < min_res[r]
            })
            nodes_sink.update({
                path_t[key][-2]: (val, r)
                for key, val in length_t.items()
                if val > max_res[r] or val < min_res[r]
            })
        except IndexError:  # No nodes violate resource limits
            pass

    nodes_source = {}
    nodes_sink = {}
    n_nodes = len(G.nodes())
    # Map function for each resource
    list(map(_check_resource, range(0, G.graph['n_res'])))
    if nodes_source and nodes_sink:  # if there are nodes to remove
        # Filter out source or sink
        nodes_to_remove = [
            node for node in G.nodes() if node in nodes_source and nodes_sink
        ]
        if "Source" in nodes_to_remove or "Sink" in nodes_to_remove:
            if "Sink" in nodes_source:
                unreachable_res = nodes_source["Sink"][1]
            elif "Sink" in nodes_sink:
                unreachable_res = nodes_sink["Sink"][1]
            elif "Source" in nodes_sink:
                unreachable_res = nodes_sink["Source"][1]
            elif "Source" in nodes_source:
                unreachable_res = nodes_source["Source"][1]
            raise Exception(
                "Sink not reachable for resource {}".format(unreachable_res))
        G.remove_nodes_from(nodes_to_remove)
        log.info("Removed {}/{} nodes".format(len(nodes_to_remove), n_nodes))
    return G


def check_and_preprocess(preprocess,
                         G,
                         max_res=None,
                         min_res=None,
                         REF_forward=None,
                         REF_backward=None,
                         direction=None,
                         algorithm=None):
    """
    Checks whether inputs and the graph are of the appropriate types and
    have the required properties.
    For non-specified REFs, removes nodes that cannot be reached due to
    resource limits.

    Parameters
    ----------
    preprocess : bool
        enables preprocessing routine.

    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats, optional
        :math:`[L, M_1, M_2, ..., M_{n\_res}]`
        upper bound for resource usage.
        We must have ``len(max_res)`` :math:`\geq 2`

    min_res : list of floats, optional
        :math:`[U, L_1, L_2, ..., L_{nres}]` lower bounds for resource usage.
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    REF_forward, REF_backward : function, optional
        Custom resource extension function. See `REFs`_ for more details.
        Default: additive, subtractive.

    direction : string, optional
        preferred search direction. Either 'both','forward', or, 'backward'.
        Default : 'both'.

    .. _REFs : https://cspy.readthedocs.io/en/latest/how_to.html#refs

    :return: If ``preprocess``, returns preprocessed graph ``G`` if no
        exceptions are raised, otherwise doesn't return anything.

    :raises: Raises exceptions if incorrect input is given.
        If multiple exceptions are raised, an exception with a list of
        exceptions is raised.
    """
    check(G, max_res, min_res, direction, algorithm)
    if REF_forward or REF_backward:
        # Cannot apply pruning with custom REFs
        _check_REFs(REF_forward, REF_backward)
        return G
    if preprocess:
        G = prune_graph(G, max_res, min_res)
        check(G)
    return G



import types
from numpy import equal
from operator import add, sub


class Label(object):
    """Label object that allows comparison and the modelling of dominance relations

    Parameters
    -----------

    weight : float
        cumulative edge weight

    node : string
        name of last node visited

    res : list
        cumulative edge resource consumption

    path : list
        all nodes in the path
    """

    _REF_forward, _REF_backward = add, sub

    def __init__(self, weight, node, res, path):
        self.weight = weight
        self.node = node
        self.res = res
        self.path = path

    def __repr__(self):
        return str(self)

    def __str__(self):  # for printing purposes
        return "Label({0},{1},{2})".format(self.weight, self.node, self.res)

    def __lt__(self, other):
        # Less than operator for two Label objects
        return (self.weight < other.weight and all(self.res <= other.res)) or (
            self.weight <= other.weight and
            (all(self.res <= other.res) and any(self.res < other.res)))

    def __le__(self, other):
        # Less than or equal to operator for two Label objects
        return self.weight <= other.weight and all(self.res <= other.res)

    def __eq__(self, other):
        # Equality operator for two Label objects
        if other:
            return (self.weight == other.weight and
                    all(equal(self.res, other.res)) and self.node == other.node)
        else:
            return False

    def __hash__(self):
        # Redefinition of hash to avoid TypeError due to the __eq__ definition
        return id(self)

    def dominates(self, other, direction):
        # Return whether self dominates other.
        if self.node != other.node:
            raise Exception("Non-comparable labels given")
        else:
            if direction == "forward":
                return (((self.weight < other.weight and
                          all(self.res <= other.res)) or
                         (self.weight <= other.weight and
                          (all(self.res <= other.res) and
                           any(self.res < other.res)))))
            elif direction == "backward":
                return (((self.weight < other.weight and
                          all(self.res >= other.res)) or
                         (self.weight <= other.weight and
                          (all(self.res >= other.res) and
                           any(self.res > other.res)))))
            else:
                raise Exception(
                    "{} cannot be used as a direction".format(direction))

    # @staticmethod
    def get_new_label(self, edge, direction, weight, res):
        # self = copy(self)
        path = list(self.path)
        node = edge[1] if direction == "forward" else edge[0]
        if node in path:  # If node already visited.
            return None
        else:
            path.append(node)
        if direction == "forward":
            if isinstance(self._REF_forward, types.BuiltinFunctionType):
                res_new = self.res + res
            else:
                res_new = self._REF_forward(self.res, edge)
        elif direction == "backward":
            if isinstance(self._REF_backward, types.BuiltinFunctionType):
                res_new = self.res - res
            else:
                res_new = self._REF_backward(self.res, edge)
        else:
            raise Exception(
                "{} cannot be used as a direction".format(direction))
        _new_label = Label(weight + self.weight, node, res_new, path)
        if _new_label == self:
            # If resulting label is the same
            return None
        return _new_label

    def feasibility_check(self, max_res, min_res):
        return all(max_res >= self.res) and all(min_res <= self.res)



from __future__ import absolute_import
from __future__ import print_function

from numpy import array
from networkx import astar_path, NetworkXException
from cspy.algorithms.path import Path
from cspy.preprocessing import check_and_preprocess


class Tabu:
    """
    Simple Tabu-esque algorithm for the (resource) constrained shortest
    path problem.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage (including initial forward stopping point).

    min_res : list of floats
        :math:`[L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage (including initial backward stopping point).
        We must have ``len(min_res)`` :math:`=` ``len(max_res)``.

    REF : function, optional
        Custom resource extension function. See `REFs`_ for more details.
        Default : additive.

    preprocess : bool, optional
        enables preprocessing routine. Default : False.

    .. _REFs : https://cspy.readthedocs.io/en/latest/how_to.html#refs

    Returns
    -------
    path : list
        nodes in shortest path obtained.

    Notes
    -----
    The input graph must have a ``n_res`` attribute.
    The edges in the graph must all have a ``res_cost`` attribute.
    See `Using cspy`_

    .. _Using cspy: https://cspy.readthedocs.io/en/latest/how_to.html

    Example
    -------
    To run the algorithm, create a :class:`Tabu` instance and call `run`.

    .. code-block:: python

        >>> from cspy import Tabu
        >>> from networkx import DiGraph
        >>> from numpy import array
        >>> G = DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('Source', 'B', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('A', 'C', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('B', 'C', res_cost=array([2, 1]), weight=-1)
        >>> G.add_edge('C', 'D', res_cost=array([1, 1]), weight=-1)
        >>> G.add_edge('D', 'E', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('D', 'F', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('F', 'Sink', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('E', 'Sink', res_cost=array([1, 1]), weight=1)
        >>> max_res, min_res = [5, 5], [0, 0]
        >>> path = Tabu(G, max_res, min_res).run()
        >>> print(path)
        ['Source', 'A', 'C', 'D', 'E', 'Sink']

    """

    def __init__(self, G, max_res, min_res, REF=None, preprocess=False):
        # Check input graph and parameters
        self.G = check_and_preprocess(preprocess, G, max_res, min_res, REF)
        # Input parameters
        self.max_res = max_res
        self.min_res = min_res
        # Algorithm specific parameters
        self.iteration = 0
        self.current_path = list()
        self.best_path = list()
        self.stop = False
        self.neighbour = 'Source'
        self.neighbourhood = list()
        self.tabu_edge = None
        self.edges_to_check = dict(self.G.edges())

        if REF:
            Path._REF = REF

    def run(self):
        """
        Runs Tabu Search with resource constraints for the input graph
        """
        while self.stop is False:
            self.algorithm()
            self.iteration += 1

        if self.best_path:
            return self.best_path
        else:
            raise Exception("No resource feasible path has been found")

    def algorithm(self):
        path = []
        try:
            path = astar_path(self.G,
                              self.neighbour,
                              'Sink',
                              heuristic=self._heuristic)
        except NetworkXException:
            pass
        if path:
            self._update_path(path)
            edge_or_true = Path(self.G, self.current_path, self.max_res,
                                self.min_res)._check_feasibility()
            if edge_or_true is True:
                self.stop = True
                self.best_path = self.current_path
            else:
                self._get_neighbour(edge_or_true)
        else:
            self._get_neighbour(self.tabu_edge)

    def _update_path(self, path):
        # Joins path using previous path and [neighbour, ..., sink] path
        if self.iteration == 0:
            self.current_path = path
        if self.neighbour in self.current_path:
            self.current_path = list(node for node in self.current_path if (
                node != self.neighbour and self.current_path.index(
                    node) < self.current_path.index(self.neighbour))) + path
        else:
            self._merge_paths(path)

    def _merge_paths(self, path):
        branch_path = [n for n in self.current_path if n not in path]
        for node in reversed(branch_path):
            if (node, self.neighbour) in self.G.edges():
                self.current_path = list(n for n in branch_path if (
                    branch_path.index(n) <= branch_path.index(node))) + path
                break

    def _get_neighbour(self, edge=None):
        if self.edges_to_check:
            if edge and edge[:2] in self.edges_to_check:
                # If edge not already been seen
                current_edge = edge
            elif edge[0] == "Source":
                self.stop = True
                return
            else:
                current_edge = self._get_next_neighbour_edge(self.tabu_edge)
        else:
            self.stop = True
            return
        self.edges_to_check.pop(current_edge[:2], None)
        self.neighbour = current_edge[0]
        self.tabu_edge = current_edge

    def _get_next_neighbour_edge(self, edge):
        # Retrieves the edge adjacent to node with the least weight
        if not self.neighbourhood:
            node = edge[0]
            if node == "Source":
                self.stop = True
                return edge
            self.neighbourhood = list(e for e in self.G.edges(
                self.G.nbunch_iter([node] + list(self.G.predecessors(node))),
                data=True) if e[1] == node and e != edge)
            self.neighbourhood.sort(key=lambda x: x[2]['weight'])
        next_edge = self.neighbourhood[-1]
        self.neighbourhood.pop(-1)
        return next_edge

    def _heuristic(self, i, j):
        # Given a node pair returns a weight to apply
        if (i, j) == self.tabu_edge:
            return 1e7
        else:
            return 0



from __future__ import absolute_import
from __future__ import print_function

import logging
from numpy import array
from networkx import astar_path, NetworkXException
from cspy.algorithms.path import Path
from cspy.preprocessing import check_and_preprocess

log = logging.getLogger(__name__)


class GreedyElim:
    """
    Simple Greedy elimination algorithm for the (resource) constrained shortest
    path problem. The algorithms solves a standard shortest path problem and
    eliminates resource infeasible edges iteratively until a resource feasible
    path is found.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage (including initial forward stopping point).

    min_res : list of floats
        :math:`[L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage (including initial backward stopping point).
        We must have ``len(min_res)`` :math:`=` ``len(max_res)``.

    REF : function, optional
        Custom resource extension function. See `REFs`_ for more details.
        Default : additive.

    preprocess : bool, optional
        enables preprocessing routine. Default : False.

    return_G : bool, optional
        whether or not you'd like the resulting graph returned

    .. _REFs : https://cspy.readthedocs.io/en/latest/how_to.html#refs

    Returns
    -------
    path : list
        nodes in shortest path obtained.

    Raises
    ------
    Exception
        if no resource feasible path is found

    Notes
    -----
    The input graph must have a ``n_res`` attribute.
    The edges in the graph must all have a ``res_cost`` attribute.
    See `Using cspy`_

    .. _Using cspy: https://cspy.readthedocs.io/en/latest/how_to.html


    Example
    -------
    To run the algorithm, create a :class:`GreedyElim` instance and call `run`.

    .. code-block:: python

        >>> from cspy import GreedyElim
        >>> from networkx import DiGraph
        >>> from numpy import array
        >>> G = DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('Source', 'B', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('A', 'C', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('B', 'C', res_cost=array([2, 1]), weight=-1)
        >>> G.add_edge('C', 'D', res_cost=array([1, 1]), weight=-1)
        >>> G.add_edge('D', 'E', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('D', 'F', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('F', 'Sink', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('E', 'Sink', res_cost=array([1, 1]), weight=1)
        >>> max_res, min_res = [5, 5], [0, 0]
        >>> path = GreedyElim(G, max_res, min_res).run()
        >>> print(path)
        ['Source', 'A', 'C', 'D', 'E', 'Sink']

    """

    def __init__(self,
                 G,
                 max_res,
                 min_res,
                 REF=None,
                 preprocess=False,
                 return_G=False):
        # Check input graph and parameters
        self.G = check_and_preprocess(preprocess, G, max_res, min_res, REF)
        # Input parameters
        self.max_res = max_res
        self.min_res = min_res
        self.return_G = return_G
        # Algorithm specific parameters
        self.it = 0
        self.path = []
        self.stop = False
        self.predecessor_edges = []
        self.last_edge_removed = None
        self.edges_to_remove = dict(self.G.edges())

        if REF:
            Path._REF = REF

    def run(self):
        while self.stop is False:
            self.algorithm()
            self.it += 1

        if self.path:
            if self.return_G:
                return self.G, self.path
            else:
                return self.path
        else:
            raise Exception("No resource feasible path has been found")

    def algorithm(self):
        path = []
        try:
            path = astar_path(self.G, 'Source', 'Sink')
        except NetworkXException:
            pass
        if path:
            edge_or_true = Path(self.G, path, self.max_res,
                                self.min_res)._check_feasibility()
            if edge_or_true is True:
                self.path = path
                self.stop = True
            else:
                self._update_graph(edge_or_true)
        else:
            self.G.add_edge(*self.last_edge_removed[:2],
                            res_cost=self.last_edge_removed[2]['res_cost'],
                            weight=self.last_edge_removed[2]['weight'])
            self._update_graph(
                self._get_predecessor_edges(self.last_edge_removed))

    def _update_graph(self, edge=None):
        self.G.remove_edge(*edge[:2])
        self.last_edge_removed = edge

    def _get_predecessor_edges(self, edge):
        if not self.predecessor_edges:
            node = edge[0]
            if node == "Source":
                self.stop = True
                return edge
            self.predecessor_edges = [
                e for e in self.G.edges(self.G.nbunch_iter(
                    [node] + list(self.G.predecessors(node))),
                                        data=True) if e[1] == node and e != edge
            ]
            self.predecessor_edges.sort(key=lambda x: x[2]['weight'])
        next_edge = self.predecessor_edges[-1]
        self.predecessor_edges.pop(-1)
        return next_edge

    @staticmethod
    def _edge_extract(edge):
        return array(edge[2]['res_cost'])



from __future__ import absolute_import

from itertools import repeat
from logging import getLogger
from collections import OrderedDict, deque

from numpy import zeros
from numpy.random import RandomState

# Local module imports
from cspy.algorithms.label import Label
from cspy.preprocessing import check_and_preprocess

log = getLogger(__name__)


class BiDirectional:
    """
    Implementation of the bidirectional labeling algorithm with dynamic
    half-way point (`Tilk 2017`_).
    Depending on the range of values for U, L, we get
    four different algorithms. See self.name_algorithm and Notes.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[H_F, M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage (including initial forward stopping point).
        We must have ``len(max_res)`` :math:`\geq 2`.

    min_res : list of floats
        :math:`[H_B, L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage (including initial backward stopping point).
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    REF_forward REF_backward : function, optional
        Custom resource extension function. See `REFs`_ for more details.
        Default : additive, subtractive.

    preprocess : bool, optional
        enables preprocessing routine. Default : False.

    direction : string, optional
        preferred search direction.
        Either "both", "forward", or, "backward". Default : "both".

    seed : None or int or numpy.random.RandomState instance, optional
        seed for PSOLGENT class. Default : None (which gives a single value
        numpy.random.RandomState).

    .. _REFs : https://cspy.readthedocs.io/en/latest/how_to.html#refs

    Returns
    -------
    list
        nodes in shortest path obtained.

    Notes
    -----
    The input graph must have a ``n_res`` attribute which must be
    :math:`\geq 2`. The edges in the graph must all have a ``res_cost``
    attribute.

    According to the inputs, four different algorithms can be used.
    If you'd like to check which algorithm is running given the resource
    limits, call the method :func:`BiDirectional.name_algorithm()`
    for a log with the classification.

    - ``direction`` = "forward": Monodirectional forward labeling algorithm
    - :math:`H_F == H_B`: Bidirectional labeling algorithm with static halfway point.
    - ``direction`` = "backward": Monodirectional backward labeling algorithm
    - :math:`H_F > H_B`: Bidirectional labeling algorithm with dynamic halfway point.
    - :math:`H_F < H_B`: The algorithm won't go anywhere!

    Example
    -------
    To run the algorithm, create a :class:`BiDirectional` instance and call
    ``run``.

    .. code-block:: python

        >>> from cspy import BiDirectional
        >>> from networkx import DiGraph
        >>> from numpy import array
        >>> G = DiGraph(directed=True, n_res=2)
        >>> G.add_edge("Source", "A", res_cost=array([1, 2]), weight=0)
        >>> G.add_edge("A", "B", res_cost=array([1, 0.3]), weight=0)
        >>> G.add_edge("A", "C", res_cost=array([1, 0.1]), weight=0)
        >>> G.add_edge("B", "C", res_cost=array([1, 3]), weight=-10)
        >>> G.add_edge("B", "Sink", res_cost=array([1, 2]), weight=10)
        >>> G.add_edge("C", "Sink", res_cost=array([1, 10]), weight=0)
        >>> max_res, min_res = [4, 20], [1, 0]
        >>> path = BiDirectional(G, max_res, min_res, direction="both").run()
        >>> print(path)
        ["Source", "A", "B", "C", "Sink"]

    .. _Tilk 2017: https://www.sciencedirect.com/science/article/pii/S0377221717302035
    """

    def __init__(self,
                 G,
                 max_res,
                 min_res,
                 REF_forward=None,
                 REF_backward=None,
                 preprocess=False,
                 direction="both",
                 seed=None):
        # Check inputs and preprocess G unless option disabled
        self.G = check_and_preprocess(preprocess, G, max_res, min_res,
                                      REF_forward, REF_backward, direction,
                                      __name__)
        self.direc_in = direction
        self.max_res, self.min_res = max_res.copy(), min_res.copy()
        # Algorithm specific parameters #
        # Current forward and backward labels
        self.currentLabel = OrderedDict({
            "forward":
            Label(0, "Source", zeros(G.graph["n_res"]), ["Source"]),
            "backward":
            Label(0, "Sink", max_res, ["Sink"])
        })
        # Unprocessed labels dict (both directions)
        self.unprocessedLabels = OrderedDict({
            "forward": OrderedDict({}),
            "backward": OrderedDict({})
        })
        # Processed labels dict
        self.processedLabels = OrderedDict({
            "forward": deque(),
            "backward": deque()
        })
        # Final labels dict.
        self.finalLabel = OrderedDict({
            "forward": self.currentLabel["forward"],
            "backward": self.currentLabel["backward"]
        })

        # If given, set REFs for dominance relations and feasibility checks
        if REF_forward and REF_backward:
            Label._REF_forward = REF_forward
            Label._REF_backward = REF_backward
        # Init with seed if given
        if seed is None:
            self.random_state = RandomState()
        elif isinstance(seed, int):
            self.random_state = RandomState(seed)
        elif isinstance(seed, RandomState):
            self.random_state = seed
        else:
            raise Exception(
                '{} cannot be used to seed numpy.random.RandomState'.format(
                    seed))

    def run(self):
        while self.currentLabel["forward"] or self.currentLabel["backward"]:
            direc = self._get_direction()
            if direc:
                self._algorithm(direc)
                self._check_dominance(direc)
            elif not direc or self._terminate(direc):
                break
        # Otherwise return either path
        return self._join_paths()

    ###########################
    # Classify Algorithm Type #
    ###########################
    def name_algorithm(self):
        """
        Determine which algorithm is running.
        """
        HF, HB = self.max_res[0], self.min_res[0]
        if self.direc_in == "forward":
            log.info("Monodirectional forward labeling algorithm")
        elif self.direc_in == "backward":
            log.info("Monodirectional backward labeling algorithm")
        elif HF > HB:
            log.info("Bidirectional labeling algorithm with" +
                     " dynamic halfway point")
        else:
            log.info("The algorithm can't move in either direction!")

    #############
    # DIRECTION #
    #############
    def _get_direction(self):
        if self.direc_in == "both":
            if (self.currentLabel["forward"] and
                    not self.currentLabel["backward"]):
                return "forward"
            elif (not self.currentLabel["forward"] and
                  self.currentLabel["backward"]):
                return "backward"
            elif (self.currentLabel["forward"] and
                  self.currentLabel["backward"]):
                return self.random_state.choice(["forward", "backward"])
            else:  # if both are empty
                return
        else:
            if (not self.currentLabel["forward"] and
                    not self.currentLabel["backward"]):
                return
            elif not self.currentLabel[self.direc_in]:
                return
            else:
                return self.direc_in

    #############
    # ALGORITHM #
    #############
    def _algorithm(self, direc):
        if direc == "forward":  # forward
            idx = 0  # index for head node
            # Update backwards half-way point
            self.min_res[0] = max(
                self.min_res[0],
                min(self.currentLabel[direc].res[0], self.max_res[0]))
        else:  # backward
            idx = 1  # index for tail node
            # Update forwards half-way point
            self.max_res[0] = min(
                self.max_res[0],
                max(self.currentLabel[direc].res[0], self.min_res[0]))
        # Select edges with the same head/tail node as the current label node.
        edges = deque(e for e in self.G.edges(data=True)
                      if e[idx] == self.currentLabel[direc].node)
        # If Label not been seen before, initialise a list
        if (self.currentLabel[direc] not in self.unprocessedLabels[direc] and
                self.currentLabel[direc] not in self.processedLabels[direc]):
            self.unprocessedLabels[direc][self.currentLabel[direc]] = deque()
        # Propagate current label along all suitable edges in current direction
        list(map(self._propagate_label, edges, repeat(direc, len(edges))))
        # Extend label
        next_label = self._get_next_label(direc)
        if next_label in self.processedLabels[direc]:
            next_label = self._get_next_label(direc, next_label)
        # Update current label
        self.currentLabel[direc] = next_label

    def _propagate_label(self, edge, direc):
        # Label propagation #
        weight, res_cost = edge[2]["weight"], edge[2]["res_cost"]
        # Get new label from current Label
        new_label = self.currentLabel[direc].get_new_label(
            edge, direc, weight, res_cost)
        if new_label and new_label.feasibility_check(self.max_res,
                                                     self.min_res):
            # If label doesn't exist and is resource feasible
            self.unprocessedLabels[direc][self.currentLabel[direc]].append(
                new_label)

    def _get_next_label(self, direc, exclude_label=None):
        # Label Extension #
        try:
            # Try to get labels create from the current one.
            _labels = list(lab for lab in self.unprocessedLabels[direc][
                self.currentLabel[direc]]
                           if (lab not in self.processedLabels[direc] and
                               lab != exclude_label))
            return min(_labels, key=lambda x: x.weight)
        except ValueError:
            # No more keys to be processed under current label
            self.processedLabels[direc].append(self.currentLabel[direc])
            _labels = list(lab for lab in self.unprocessedLabels[direc].keys()
                           if (lab not in self.processedLabels[direc] and
                               lab != exclude_label))
            if _labels:
                return min(_labels, key=lambda x: x.weight)
            else:
                return None
        except KeyError:
            # No more keys to be processed.
            return None

    def _save_current_best_label(self, direc):
        try:
            if self.currentLabel[direc].dominates(self.finalLabel[direc],
                                                  direc):
                log.debug("Saving {} as best, with path {}".format(
                    self.currentLabel[direc], self.currentLabel[direc].path))
                self.finalLabel[direc] = self.currentLabel[direc]
        except Exception:
            # Labels are not comparable
            if (direc == "forward" and
                ((self.currentLabel[direc].path[-1] == "Sink" or
                  self.finalLabel[direc].node == "Source") or
                 ((self.currentLabel[direc].node not in
                   self.finalLabel[direc].path) and
                  (self.currentLabel[direc].weight <=
                   self.finalLabel[direc].weight)))):
                log.debug("Saving {} as best, with path {}".format(
                    self.currentLabel[direc], self.currentLabel[direc].path))
                self.finalLabel[direc] = self.currentLabel[direc]
            elif (direc == "backward" and
                  ((self.currentLabel[direc].path[-1] == "Source" or
                    self.finalLabel[direc].node == "Sink") or
                   ((self.currentLabel[direc].node not in
                     self.finalLabel[direc].path) and
                    (self.currentLabel[direc].weight <=
                     self.finalLabel[direc].weight)))):
                log.debug("Saving {} as best, with path {}".format(
                    self.currentLabel[direc], self.currentLabel[direc].path))
                self.finalLabel[direc] = self.currentLabel[direc]

    #############
    # DOMINANCE #
    #############
    def _check_dominance(self, direc):
        """
        For all labels, check if it is dominated, or itself dominates other
        labels. If this is found to be the case, the dominated label is
        removed.
        """
        if self.currentLabel[direc]:
            keys_to_pop = deque()
            all_labels = deque(
                lab for lab in self.unprocessedLabels[direc].keys()
                if lab.node == self.currentLabel[direc].node and
                lab != self.currentLabel[direc])
            all_labels.extend(
                lab for k, v in self.unprocessedLabels[direc].items()
                for lab in v if lab.node == self.currentLabel[direc].node and
                lab != self.currentLabel[direc])
            keys_to_pop.extend(
                lab for lab in all_labels
                if self.currentLabel[direc].dominates(lab, direc))
            keys_to_pop.extend([
                lab for lab in self.processedLabels[direc]
                if lab != self.finalLabel[direc]
            ])

            if any(
                    lab.dominates(self.currentLabel[direc], direc)
                    for lab in all_labels):
                keys_to_pop.append(self.currentLabel[direc])
            else:
                self._save_current_best_label(direc)
            self._remove_unprocessed_labels(keys_to_pop, direc)

    def _remove_unprocessed_labels(self, keys_to_pop, direc):
        # Remove all processed labels from unprocessed dict
        for key_to_pop in list(set(keys_to_pop)):
            if key_to_pop in self.unprocessedLabels[direc]:
                log.debug("Key {} removed".format(key_to_pop))
                del self.unprocessedLabels[direc][key_to_pop]
            for k, sub_dict in self.unprocessedLabels[direc].items():
                if key_to_pop in sub_dict:
                    _idx = sub_dict.index(key_to_pop)
                    log.debug("Key {} removed from sub_dict".format(key_to_pop))
                    del self.unprocessedLabels[direc][k][_idx]

    ###############
    # TERMINATION #
    ###############
    def _terminate(self, direc):
        if self.direc_in == "both":
            if (not self.unprocessedLabels["forward"] or
                    not self.unprocessedLabels["backward"]):
                return True
        else:
            if (self.direc_in == "forward" and
                    self.finalLabel["forward"].path[-1] == "Sink" and
                    not self.unprocessedLabels["forward"]):
                return True
            elif (self.direc_in == "backward" and
                  self.finalLabel["backward"].path[-1] == "Source" and
                  not self.unprocessedLabels["backward"]):
                return True

    #################
    # PATH CHECKING #
    #################
    def _join_paths(self):
        # check if paths are eligible to be joined
        if self.direc_in == "both":
            return self._check_paths()
        else:
            if self.direc_in == "backward":
                self.finalLabel[self.direc_in].path.reverse()
            return self.finalLabel[self.direc_in].path

    def _check_paths(self):
        # Reverse backward path
        self.finalLabel["backward"].path.reverse()
        if (self.finalLabel["forward"].path[-1] == "Sink" and
                self.finalLabel["backward"].path[0] != "Source"):
            # if only forward path
            return self.finalLabel["forward"].path
        elif (self.finalLabel["backward"].path[0] == "Source" and
              self.finalLabel["forward"].path[-1] != "Sink"):
            # if only backward path
            return self.finalLabel["backward"].path
        elif (self.finalLabel["backward"].path[0] == "Source" and
              self.finalLabel["forward"].path[-1] == "Sink"):
            # if both paths
            if self.finalLabel["forward"].weight < self.finalLabel[
                    "backward"].weight:
                # if forward path has a lower weight
                return self.finalLabel["forward"].path
            elif self.finalLabel["backward"].weight < self.finalLabel[
                    "forward"].weight:
                # if backward path has a lower weight
                return self.finalLabel["backward"].path
            else:
                # Otherwise (equal weight) return either path
                return (self.finalLabel["forward"].path
                        if self.random_state.random_sample() < 0.5 else
                        self.finalLabel["backward"].path)
        else:
            # if combination of the two is required
            return list(
                OrderedDict.fromkeys(self.finalLabel["forward"].path +
                                     self.finalLabel["backward"].path))



from setuptools import setup, find_packages

setup(
    name="cspy",
    version="0.0.11",
    description=
    "A collection of algorithms for the (Resource) Constrained Shortest Path Problem",
    license="MIT",
    author="David Torres",
    author_email="d.torressanchez@lancs.ac.uk",
    keywords=[
        "shortest path", "resource constrained shortest path",
        "bidirectional algorithm", "tabu search"
    ],
    long_description_content_type="text/x-rst",
    long_description=open("README.rst", "r").read(),
    url="https://github.com/torressa/cspy",
    packages=find_packages(),
    install_requires=['networkx', 'numpy'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from networkx import (DiGraph, NetworkXException, NetworkXUnbounded, has_path,
                      negative_edge_cycle)
from numpy import ndarray

__all__ = [
    "_check_res", "_check_direction", "_check_graph_attr", "_check_edge_attr",
    "_check_path", "_check_REFs"
]


def _check_res(G, max_res, min_res, direction, algorithm):
    if isinstance(max_res, list) and isinstance(min_res, list):
        if len(max_res) == len(min_res):
            if (algorithm and 'bidirectional' in algorithm and
                    len(max_res) < 2):
                raise TypeError("Resources must be of length >= 2")
            if (all(isinstance(i, (float, int)) for i in max_res) and
                    all(isinstance(i, (float, int)) for i in min_res)):
                pass
            else:
                raise TypeError("Elements of input lists must be numbers")
        else:
            raise TypeError("Input lists have to be equal length")
    else:
        raise TypeError("Inputs have to be lists with length >= 2")


def _check_direction(G, max_res, min_res, direction, algorithm):
    if direction not in ['forward', 'backward', 'both']:
        raise TypeError(
            "Input direction has to be 'forward', 'backward', or 'both'")


def _check_graph_attr(G, max_res, min_res, direction, algorithm):
    """Checks whether input graph has n_res attribute"""
    if isinstance(G, DiGraph):
        if 'n_res' not in G.graph:
            raise TypeError("Input graph must have 'n_res' attribute.")
    else:
        raise TypeError("Input must be a nx.Digraph()")


def _check_edge_attr(G, max_res, min_res, direction, algorithm):
    """Checks whether edges in input graph have res_cost attribute"""
    if not all('res_cost' in edge[2] for edge in G.edges(data=True)):
        raise TypeError(
            "Input graph must have edges with 'res_cost' attribute.")
    if not all(
            len(edge[2]['res_cost']) == G.graph['n_res']
            for edge in G.edges(data=True)):
        raise TypeError(
            "Edges must have 'res_cost' attribute with length equal to 'n_res'")
    if not all(
            len(edge[2]['res_cost']) == len(max_res) == len(min_res)
            for edge in G.edges(data=True)):
        raise TypeError(
            "Edges must have 'res_cost' attribute with length equal to" +
            "'min_res' == 'max_res")
    if not all(
            isinstance(edge[2]['res_cost'], ndarray)
            for edge in G.edges(data=True)):
        raise TypeError("The edge 'res_cost' attribute must be a numpy.array")


def _check_path(G, max_res, min_res, direction, algorithm):
    """Checks whether a 'Source' -> 'Sink' path exists and if there are
    negative edge cycles in the graph.
    Also covers nodes missing and other standard networkx exceptions."""
    try:
        if not has_path(G, 'Source', 'Sink'):
            raise NetworkXException("Disconnected Graph")
        if negative_edge_cycle(G):
            raise NetworkXException("A negative cost cycle was found.")
    except NetworkXException as e:
        raise Exception("An error occurred: {}".format(e))


def _check_REFs(REF_forward, REF_backward):
    if (REF_forward and not callable(REF_forward)) or (
            REF_backward and
            not callable(REF_backward)) or (REF_forward and REF_backward and
                                            not callable(REF_forward) and
                                            not callable(REF_backward)):
        raise TypeError("REF functions must be callable")
